// Copyright (c) 2002 Tom Barry.  All rights reserved.
//		trbarry@trbarry.com
//	modified by Foxyshadis
//		foxyshadis@hotmail.com
// Requires Avisynth source code to compile for Avisynth
// Avisynth Copyright 2000 Ben Rudiak-Gould.
//      http://www.math.berkeley.edu/~benrg/avisynth.html
/////////////////////////////////////////////////////////////////////////////
//
//	This file is subject to the terms of the GNU General Public License as
//	published by the Free Software Foundation.  A copy of this license is
//	included with this software distribution in the file COPYING.  If you
//	do not have a copy, you may obtain a copy by writing to the Free
//	Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//
//	This software is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details
//  
//  Also, this program is "Philanthropy-Ware".  That is, if you like it and 
//  feel the need to reward or inspire the author then please feel free (but
//  not obligated) to consider joining or donating to the Electronic Frontier
//  Foundation. This will help keep cyber space free of barbed wire and bullsh*t.  
//
/////////////////////////////////////////////////////////////////////////////
// Change Log
//
// Date          Version  Developer      Changes
//

// 07 May 2003   1.0.0.0  Tom Barry      New Release
// 01 Jun 2006   1.1.0.0  Foxyshadis     Chroma noise, constant seed
// 06 Jun 2006   1.2.0.0  Foxyshadis     Supports YUY2, RGB. Fix cache mess.
// 10 Jun 2006   1.3.0.0  Foxyshadis     Crashfix, noisegen optimization
// 11 Nov 2006   1.4.0.0  Foxyshadis     Constant replaces seed, seed repeatable
// 07 May 2010   1.5.0.0  Foxyshadis     Limit the initial seed generation to fix memory issues.
// 13 May 2010   1.5.1.0  Firesledge     The source code compiles on Visual C++ versions older than 2008
// 26 Oct 2011   1.5.2.0  Firesledge     Removed the SSE2 requirement.
//
/////////////////////////////////////////////////////////////////////////////
// CVS Log
//
//
/////////////////////////////////////////////////////////////////////////////
#pragma warning(disable:4244) // disable conversion from double to float message

#ifdef DEBUG
#define _CRTDBG_MAP_ALLOC
#include <stdlib.h>
#include <crtdbg.h>
#endif

#define WIN32_LEAN_AND_MEAN
#define VC_EXTRALEAN
#include <windows.h>
#include "avisynth.h"
#include <math.h>
#include <time.h>

#if defined (_MSC_VER) && _MSC_VER < 1500
	typedef	unsigned char	uint8_t;
#else
	#include <stdint.h>
#endif

#define ALIGN8(x) (((x) + 7) >>3 ) << 3
#define ALIGN16(x) (((x) + 15) >>4 ) << 4

// max # of noise planes, try 3 if you want to rewrite the yv12 chroma routines
#define MAXP 2

class AddGrain : public GenericVideoFilter {
	int  Max_Frames;

	int planes;
	float variance, uvariance;
	float Hcorr, Vcorr;
	int seed,storedframes;
	bool iset, constant;
	float gset;
	long idum;
	char* pNW;			// work ptr to noise
	uint8_t* pNoiseSeeds;
	// increase this to lessen inter-frame noise coherence and increase memory
	float NRep[MAXP];
	char* pN[MAXP];
	int NPitch[MAXP], NHeight[MAXP], NSize[MAXP];

	PVideoFrame GetFramePlanar(int inFrame, IScriptEnvironment* env);
	PVideoFrame GetFrameInterleaved(int inFrame, IScriptEnvironment* env);
	BOOL UpdateFrame(int dst_pit, int row_size, BYTE* dstp, int FldHeight, int plane);

	float FastUniformRandF();
	long FastUniformRandL();
	float GaussianRand();
	float GaussianRand(float mean, float variance);
	void SetRand(int plane, int frame_number );

public:

	AddGrain(PClip, float, float, float, float,int,bool, IScriptEnvironment*);
	PVideoFrame __stdcall GetFrame(int inFrame, IScriptEnvironment* env);
	~AddGrain() 
	{
		for(int x=0;x<MAXP;x++)
			if(pN[x])
				free(pN[x]);
		if(pNoiseSeeds)
			free(pNoiseSeeds);
	}
};

AddGrain::AddGrain(PClip _child, float _variance, float _Hcorr, float _Vcorr, float _uvariance, 
		int _seed, bool _constant, IScriptEnvironment* env) : 
	GenericVideoFilter(_child), variance(_variance), Hcorr(_Hcorr), Vcorr(_Vcorr), uvariance(_uvariance),
		seed(_seed), constant(_constant)
 {
	if ((env->GetCPUFlags() & CPUF_MMX) == 0)
	{
		env->ThrowError("AddGrain: Only MMX CPU's currently supported");
	}

    if (Hcorr < 0.0 || Hcorr > 1.0 || Vcorr < 0.0 || Vcorr > 1.0)
	{
		env->ThrowError("AddGrain: HCorr & VCorr must be 0 <= x <= 1.0");
	}
	child->SetCacheHints(CACHE_NOTHING, 0);
	int x,y;
	float r;
	float LastLine[2][4096];
	float Lastr;
	float mean = 0;
	iset = false;
	if(seed<0)
		seed = (unsigned long)time(NULL);		// init random
	idum = seed;
	for(x=0;x<MAXP;x++)
		pN[x] = 0;

	// set up requirements for different colorspaces
	if(vi.IsYV12()) {
		planes = 2;
		NPitch[0] = ALIGN8(vi.width);		// luma
		NPitch[1] = ALIGN8(vi.width >> 1);		// chroma
		NHeight[0] = vi.height;
		NHeight[1] = vi.height >> 1;
	} else if(vi.IsYUY2()) {
		planes = 2;
		NPitch[0] = NPitch[1] = ALIGN8(vi.width * 2);
		NHeight[0] = NHeight[1] = vi.height;
	} else if(vi.IsRGB()) {
		planes = 1;
		NPitch[0] = ALIGN8(vi.width * vi.BytesFromPixels(1));
		NHeight[0] = vi.height;
	} else {
		env->ThrowError("AddGrain: Unknown colorspace!");
	}
	// todo: add malloc error handling
	storedframes = min(vi.num_frames,256);
	long NoiseSize = storedframes * planes * sizeof(pNoiseSeeds);
	uint8_t* pns = pNoiseSeeds = (uint8_t*) malloc(NoiseSize); 
	NRep[0] = NRep[1] = 2;
	if(constant) {
		//seed = seed % NoiseSize;
		NRep[0] = 1;
		NRep[1] = !vi.IsYV12() ? 1 : 1 + 32 / (NPitch[1] * NPitch[1]);
	}

	float pvar[] = { variance, uvariance };
	
	for(int plane=0; plane < planes; plane++)	{
		int h = NHeight[plane] * NRep[plane];
		NSize[plane] = NPitch[plane] * h + 16;	// small safety margin
		// todo: add malloc error handling
		// allocate space for noise
		pN[plane] = (char*) malloc(NSize[plane]);
		for (x=0 ; x < NPitch[plane]; x++) {
			LastLine[plane][x] = GaussianRand(mean, pvar[plane]);		// things to vertically smooth against
		}
		for (y=0; y < h; y++) {
			pNW = pN[plane] + y * NPitch[plane];
			Lastr = GaussianRand(mean, pvar[plane]);   // something to horiz smooth against
			for (x = 0; x < NPitch[plane]; x++) {
				r = GaussianRand(mean, pvar[plane]);
				Lastr = r = Lastr* Hcorr + r * (1-Hcorr);  // horizontal correlation
				LastLine[plane][x] = r = LastLine[plane][x] * Vcorr + r * (1-Vcorr); // vert corr

				// set noise block
				if(vi.IsYV12()) {
					*pNW++ = (char) r ;
				} else if(vi.IsYUY2()) {
					*pNW++ = (x++ + plane) % 2 == 0 ? (char) r : 0;
					*pNW++ = (x   + plane) % 2 == 0 ? (char) r : 0;
					// alternating pixels: plane 0 sets y0y0, plane 1 sets 0u0v
				} else if(vi.IsRGB()) {
					*pNW++ = (char) r ;		// all channels equal, no chroma grain
					*pNW++ = (char) r ; x++;
					*pNW++ = (char) r ; x++;
					if(vi.IsRGB32()) {
						*pNW++ = 0;  x++;		// alpha channel unused
					}
				}
			}
		}
		for (x=storedframes; x > 0; x--) {
			// insert seed, to keep cache happy
			*pns++ = FastUniformRandL() & 0xff;
		}
	}
}


PVideoFrame __stdcall AddGrain::GetFrame(int inFrame, IScriptEnvironment* env) {
	if(vi.IsPlanar()) {
		return GetFramePlanar(inFrame,env);
	} else {
		return GetFrameInterleaved(inFrame,env);
	}
}

PVideoFrame AddGrain::GetFramePlanar(int inFrame, IScriptEnvironment* env) {
	PVideoFrame src = child->GetFrame(inFrame, env);
	env->MakeWritable(&src);
	BYTE* dstpY;
	BYTE* dstpV;   
	BYTE* dstpU;
	int dst_pitchY;
	int dst_pitchUV;
	int row_sizeY;
	int row_sizeUV;
	int heightY;
	int heightUV;

	dstpY = src->GetWritePtr(PLANAR_Y);
	dstpV = src->GetWritePtr(PLANAR_V);   
	dstpU = src->GetWritePtr(PLANAR_U);

	dst_pitchY = src->GetPitch(PLANAR_Y);
	dst_pitchUV = src->GetPitch(PLANAR_U);

	row_sizeY = src->GetRowSize(PLANAR_Y);   // Could also be PLANAR_Y_ALIGNED which would return a mod16 row_size   
	row_sizeUV = src->GetRowSize(PLANAR_U);  // Could also be PLANAR_U_ALIGNED which would return a mod8 row_size   

	heightY = src->GetHeight(PLANAR_Y);
	heightUV = src->GetHeight(PLANAR_U);

	if(variance > 0.0f) {
		SetRand(0,inFrame);	// seeds randomness w/ plane & frame
		UpdateFrame(dst_pitchY,	 row_sizeY,  dstpY, heightY, 0);
	}

	if(uvariance > 0.0f) {
		SetRand(1,inFrame);
		UpdateFrame(dst_pitchUV, row_sizeUV, dstpU, heightUV, 1);

		SetRand(2,inFrame);
		UpdateFrame(dst_pitchUV, row_sizeUV, dstpV, heightUV, 1);
	}

	__asm emms;
	return src;
}

PVideoFrame  AddGrain::GetFrameInterleaved(int inFrame, IScriptEnvironment* env) {
	PVideoFrame src = child->GetFrame(inFrame, env);
	env->MakeWritable(&src);
	BYTE* dstp;
	int dst_pitch;
	int row_size;
	int height;

	dstp = src->GetWritePtr();
	dst_pitch = src->GetPitch();
	row_size = src->GetRowSize();
	height = src->GetHeight();

	if(variance > 0.0f) {
		SetRand(0,inFrame);
		UpdateFrame(dst_pitch, row_size, dstp, height, 0);
	}

	if(uvariance > 0.0f && vi.IsYUY2()) {
		SetRand(1,inFrame);
		UpdateFrame(dst_pitch, row_size, dstp, height, 1);
	}

	__asm emms;
	return src;
}

__int64 All128 = 0x8080808080808080;

// todo: parallelize
BOOL	AddGrain::UpdateFrame(int dst_pit, int row_size, BYTE* dstp, int FldHeight, int plane)	
{
	int ct = (row_size+7) >> 3;
	char* pNW2 = pNW;				// so VS6 inline asm can access
	int NoisePitch2 = NPitch[plane];	// "
	__asm
		{
		mov		edi, dstp			// ptr to src and dst
		mov		esi, pNW2			// our starting loc in noise
		mov		edx, FldHeight    // ctr for num lines
		movq	mm7, All128

		align	16
NoisyLineLoop:
		xor		ebx, ebx			// offset into lines
		mov     ecx, ct
		
		align	16
NoisyCharLoop:
		movq	mm0, qword ptr[edi+ebx]	
		psubb	mm0, mm7				// turn into signed number
		movq	mm1, qword ptr[esi+ebx] // >>>>>>>> for debug only
		paddsb  mm0, qword ptr[esi+ebx] // add in our noise, saturate overflow
		paddb	mm0, mm7				// turn back into unsigned number
		movq	qword ptr[edi+ebx], mm0 // and put it back in place
		add		ebx, 8					// bump offset to next qword
		loop	NoisyCharLoop			// loop for next byte

		add		edi, dst_pit
		add		esi, NoisePitch2
		dec		edx
		jnz		NoisyLineLoop			// loop for next line
		
		emms
	}
		return 0;
}


// very fast & reasonably random
inline float AddGrain::FastUniformRandF() {
	// Works with 32-bit IEEE floating point only!

	FastUniformRandL();
	unsigned long itemp = 0x3f800000 | (0x007fffff & idum);
	return (*(float*)&itemp) - 1.0f;
}

inline long AddGrain::FastUniformRandL() {
	return idum = 1664525L * idum + 1013904223L;
}

float AddGrain::GaussianRand()
{
	float fac, rsq, v1, v2;

	// return saved second
	if(iset)
	{
		iset = false;
		return gset;
	}

	do
	{
		v1 = 2.0f * FastUniformRandF() - 1.0f;
		v2 = 2.0f * FastUniformRandF() - 1.0f;
		rsq = v1 * v1 + v2 * v2;
	} while(rsq >= 1.0f || 0.0f == rsq);

	fac = (float)sqrt(-2.0f * (float)log(rsq) / rsq);

	// function generates two values every iteration, so save one for later
	gset = v1 * fac;
	iset = true;

	return v2 * fac;
}

float AddGrain::GaussianRand(float mean, float variance)
{
	if(variance == 0.0f) return mean;
	return GaussianRand() * (float)sqrt(variance) + mean;
}

void AddGrain::SetRand(int plane, int frame_number = -1) {
	int NoiseOffs;
	if(constant)
		// force noise to be identical every frame
		if(plane==2) {
			plane--;
			NoiseOffs = 32;
		} else {
			NoiseOffs = 0;
		}
	else {
		if(frame_number >= 0) {
			// pull seed back out, to keep cache happy
			int p0 = pNoiseSeeds[frame_number % storedframes];
			int p1 = (pNoiseSeeds + storedframes)[frame_number % storedframes];
			int p2 = p0 ^ p1;		// the trick to needing only 2 planes for yv12 ^.~
			switch(plane){
			case 0:
				idum = p0; break;
			case 1:
				idum = p1; break;
			case 2:
				idum = p2; plane--; break;
			}
		}
		// start noise at random qword in top half of noise area
		NoiseOffs = ((int) (FastUniformRandF() * NSize[plane]/MAXP)) & 0xfffffff8;
	}
	pNW = pN[plane] + NoiseOffs;
}

AVSValue __cdecl Create_AddGrain(AVSValue args, void* user_data,
IScriptEnvironment* env) 
{
	__asm emms;
	enum { CLIP, VARIANCE, HCORR, VCORR, UVARIANCE, SEED, CONSTANT };
	return new AddGrain(args[CLIP].AsClip(), args[VARIANCE].AsFloat(1.0), args[HCORR].AsFloat(0.0),
		args[VCORR].AsFloat(0.0), args[UVARIANCE].AsFloat(0.0), args[SEED].AsInt(-1), args[CONSTANT].AsBool(false),  env);
}

AVSValue __cdecl Create_AddGrainC(AVSValue args, void* user_data,
IScriptEnvironment* env) 
{
	__asm emms;
	enum { CLIP, VARIANCE, UVARIANCE, HCORR, VCORR, SEED, CONSTANT };
	return new AddGrain(args[CLIP].AsClip(), args[VARIANCE].AsFloat(1.0), args[HCORR].AsFloat(0.0),
		args[VCORR].AsFloat(0.0), args[UVARIANCE].AsFloat(0.0), args[SEED].AsInt(-1), args[CONSTANT].AsBool(false),  env);
}

extern "C" __declspec(dllexport) const char* __stdcall
AvisynthPluginInit2(IScriptEnvironment* env) {
			_CrtSetDbgFlag ( _CRTDBG_ALLOC_MEM_DF | _CRTDBG_LEAK_CHECK_DF );
    env->AddFunction("AddGrain", "c[var]f[hcorr]f[vcorr]f[uvar]f[seed]i[constant]b", Create_AddGrain, 0);
    env->AddFunction("AddGrainC", "c[var]f[uvar]f[hcorr]f[vcorr]f[seed]i[constant]b", Create_AddGrainC, 0);
    return "`AddGrainC' Add some correlated color gaussian noise";
}

